#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2007-2010 Christian Dannie Storgaard
#
# AUTHOR:
# Christian Dannie Storgaard <cybolic@gmail.com>

from __future__ import print_function

import common, util, monitor, desktop, parsers
import os, sys, urllib, re, subprocess

import logging as _logging

logging = _logging.getLogger("python-wine")
debug = logging.debug
error = logging.error

def check_setup():
    """
        Check whether the Wine installation has been created.
        Returns True or False
    """
    paths = [
        # The user profile directory isn't created by Wine until an application
        # actually puts something in there, so don't look for it.
        #"%s/dosdevices/c:/windows/profiles/%s" % (common.ENV['WINEPREFIX'], common.ENV['USER']),
        "%s/dosdevices/c:/windows/system32" % common.ENV['WINEPREFIX'],
        "%s/system.reg" % common.ENV['WINEPREFIX'],
        "%s/userdef.reg" % common.ENV['WINEPREFIX'],
        "%s/user.reg" % common.ENV['WINEPREFIX'],
    ]
    return all([os.path.exists(path) for path in paths])


def winetricks_installed():
    """"Return the path to winetricks, if installed, else return False.
Also checks if winetricks is actually a shell script."""
    winetricks_path = common.which('winetricks')
    if winetricks_path:
        return winetricks_path

    winetricks_path = '%s/winetricks.sh' % common.ENV['VINEYARDPATH']

    if os.access(winetricks_path, os.R_OK) and os.path.getsize(winetricks_path):
        with open(winetricks_path, 'r') as file_obj:
            content = file_obj.read()
            if '#!' in content:
                runner = content[content.find('#!'):].split('\n')[0]
                if (
                    runner.endswith('sh') or
                    runner.endswith('bash') or
                    runner.endswith('dash')
                ):
                    return winetricks_path
    return False

def update_winetricks():
    import urllib
    winetricks_uri = 'http://winezeug.googlecode.com/svn/trunk/winetricks'

    winetricks_path = '%s/winetricks.sh' % common.ENV['VINEYARDPATH']
    urllib.urlretrieve(winetricks_uri, winetricks_path)
    return winetricks_path


def run_winetricks(arguments, shell_output=False, output_to_files=False):
    """
    Run (and possibly download winetricks first) winetricks with arguments.
    arguments has to be of type list."""

    if shell_output is True and output_to_files is True:
        raise Exception, "shell_output and output_to_files can't both be True"

    winetricks_path = winetricks_installed()
    if not winetricks_path:
        winetricks_path = update_winetricks()

    return monitor.Winetricks(
        arguments,
        name = ', '.join(arguments),
        env = common.ENV,
        executable = winetricks_path
    )

def parse_command(command, env=None):
    """Parse a command, outputting a dict containing executable, arguments,"""+\
    """environment and so on."""+\
    """Commands that have better XDG replacements are replaced."""+\
    """command can be list (preferebly) or string"""

    # If command is given as a string
    if type(command) in (str, unicode):
        if util.path_exists(command):
            command = [command]
        elif util.path_exists(util.unescape_string(command)):
            command = [util.unescape_string(command)]
        else:
            command = util.string_split(
                command,
                retain_defines=True,
                remove_escapes=True
            )
    else:
        # This is a list, meaning it has to be preparsed, don't change it
        command = command[:] # make a copy, not a link

    # These variables are reported back in a dict
    wine_prefix = None      # WINEPREFIX, will be set in environment
    executable = None       # The executable to run, usually returns wine
    internal = False        # If True, don't monitor since nothing should go wrong
    if env is None or type(env) is not dict:
        env = common.ENV.copy() # Self-explanatory - except that any 'env'-command will alter it
    run_in_path = None      # Start the command in this path
    name = None             # Name will be set to something explanatory, like "Console"

    # Remove double backslashes
    command = [ i.replace('\\\\', '\\') for i in command ]

    # Move any env definitions to the env dict
    if command[0] == 'env':
        for index, arg in enumerate(command[1:]):
            if '=' in arg:
                key = arg.split('=')[0]
                value= '='.join(arg.split('=')[1:])
                if value[0] in ('"', "'"):
                    value = value[1:-1]
                env[key] = value
            else:
                break
        command = command[1+index:]
    if command[0] == 'wine':
        del command[0]
        executable = common.ENV['WINE']
    if len(command) == 1:
        command[0] = util.string_remove_escapes(command[0])

    # Find all file and directory paths in the command
    possible_paths = []
    for index, i in enumerate(command):
        if i[1:3] == ':\\':
            possible_paths.append((index, i))
        elif i[0] == '/':
            if os.path.exists(i):
                possible_paths.append((index, i))
            elif os.path.exists(util.unescape_string(i)):
                possible_paths.append((index, util.unescape_string(i)))
                command[index] = util.unescape_string(i)
    #print("POSSIBLE PATHS:",possible_paths)

    # Get the last file argument and also get the base path for run_in_path
    if len(possible_paths):
        for index, path in possible_paths:
            # Use the first existing path as the run_in_path variable
            # This means using the path to the executable, not the argument
            if util.path_exists(path):
                run_in_path = os.path.dirname(util.wintounix(path))
                break
        last_file_arg = util.unixtowin(possible_paths[-1][1])
        last_file_arg_index = possible_paths[-1][0]
    else:
        last_file_arg = None
        last_file_arg_index = None
    print("possible paths: ",possible_paths)

    first_command = command[0].lower()
    #print("FIRST:",first_command)

    #print("LAST EXE:", last_file_arg)


    # Known, simple commands that don't require parsing
    if first_command in ('sh', 'env'):
        # Don't parse this command, it's shell magic
        executable = command[1]
        command = command[2:]
        name = "Shell command"
    elif first_command == 'winetricks':
        # Get the path for winetricks, if it exists
        winetricks = winetricks_installed()
        if winetricks:
            command[0] = winetricks
        internal = True

    # More advanced/troublesome commands
    else:
        command_lowercase = [ i.lower() for i in command ]

        # If internal command
        if (
            first_command.split('.')[-1] in common.BUILTIN_EXECUTABLES
        ) or (
            first_command[2:].startswith('\\windows\\command\\') and
            first_command.split('\\')[-1].split('.')[0] in common.BUILTIN_EXECUTABLES
        ):
            first_command_name = first_command.split('\\')[-1].split('.')[0]

            # Check for command being for a program that should be run in a command console
            if first_command_name == 'cmd':
                if len(command) > 1:
                    name = "Console: {0}".format(command[1])
                else:
                    name = "Console"
                command = [
                    'wineconsole', '--backend=user', 'cmd'
                ] + command[1:]

            elif first_command_name == 'regedit':
                name = "Regedit"
                internal = True

            # Override start.exe opening directories or URL, use xdg-open
            elif first_command_name == 'start':
                #print(command_lowercase)
                if '/unix' in command_lowercase:
                    # This is already a unix path, don't convert
                    path = command[
                        command_lowercase.index('/unix')+1
                    ]
                    path = util.wintounix(util.unixtowin(util.string_remove_escapes(path)))
                else:
                    path = util.wintounix(possible_paths[1])

                link = parsers.read_lnk_file(path)
                if link:
                    path = link

                # Only use this path if it's actually a URL or directory
                if (
                    '://' in path[:8] or
                    os.path.isdir(util.wintounix(path))
                ):
                    executable = 'xdg-open'
                    command = [path]
                    run_in_path = None
                    internal = True


    # Check for supported file types
    if last_file_arg is not None:
        if last_file_arg.endswith('.cpl'):
            if not first_command.endswith('control'):
                command = ['control'] + command[1:]
            name = "Control Panel"
            internal = True

        elif last_file_arg.endswith('.lnk'):
            # This is a Windows shell link, read the target and open that directly
            link = parsers.read_lnk_file(util.wintounix(last_file_arg))
            if link:
                # Convert link to a unix path and back to deal with Wine's
                # inability to handle Windows "wildcards" (the tilde character)
                link = util.unixtowin(util.wintounix(link))

                link_unix = util.wintounix(link)

                # If the file type is not handled by Wine normally,
                # then open it with xdg-open
                if (
                    os.path.exists(link_unix) and
                    util.get_file_mimetype(link_unix) not in common.WINDOWS_FORMATS
                ):
                    executable = 'xdg-open'
                    command = [link_unix]
                    run_in_path = None
                    internal = True
                else:
                    # Change the .lnk filename to the one it's pointing to
                    command[last_file_arg_index] = link

                    remaining_args = [
                        i.lower() for i in command
                        if (
                            i != link and
                            not i.lower().endswith('start') and
                            not i.lower().endswith('start.exe')
                        )
                    ]
                    # This is just a call to a normal file, we can handle that
                    if remaining_args == ['/unix']:
                        parsed_command = parse_command([link], env)
                        return parsed_command

        elif last_file_arg.endswith('.msi'):
            # This is a Windows Installer, run in and monitor it
            command = ['msiexec', '/i'] + command
            name = "Installer"

        elif last_file_arg.endswith('.url'):
            if last_file_arg[1:].startswith(':\\'):
                url_filename = util.wintounix(last_file_arg)
                with open(url_filename, 'r') as url_fileobj:
                    url = url_fileobj.read()
                url = re.search(r'(?im)\[InternetShortcut\]\s+URL=(.*?)\s+?$', url)
                if url:
                    #print("It checks out, run it.")
                    # This is a URL, open it in a web browser, of course without monitoring it
                    url = url.groups()[0]
                    executable = 'xdg-open'
                    command = [url]
                    run_in_path = None
                    internal = True

    for index, argument in enumerate(command):
            command[index] = util.enhance_windows_path(
                command[index]
            )

    return {
        'command': executable,
        'arguments': command,
        'internal': internal,
        'prefix': wine_prefix,
        'env': env,
        'path': run_in_path,
        'name': name
        # Debug variables
        ,'_possible_paths': possible_paths
        ,'_last_file': last_file_arg
    }


def run(command, run_in_dir=True, dont_run=False, monitor=True, shell_output=False, name=None):
    """Intelligently runs a Windows program.
This function also handles start.exe arguments (like web links), .msi installers and provides
the alias "cmd" for launching the wineconsole and "winetricks" for running, and optionally downloading
winetricks, with arguments."""

    command = parse_command(command)

    if 'VINEYARD_NO_MONITORING' in command['env']:
        monitor = False

    if monitor and shell_output:
        error("monitor and shell_output can't both be True, disabling shell_output")
        shell_output = False

    # Override any WINEPREFIX defined by command
    command['env']['WINEPREFIX'] = common.ENV['WINEPREFIX']

    # Create a new environment dict from our internal one
    env = common.copy(common.ENV)
    # Update it with the command's environment
    env.update(command['env'])

    if name is not None:
        command['name'] = name

    kwargs = {
        'env': env,
    }

    if command['command'] is None:
        command['command'] = common.ENV['WINE']

    if run_in_dir and command['path'] is not None:
        kwargs['cwd'] = command['path']


    if monitor:
        kwargs['output_to_shell'] = False
        kwargs['use_log'] = True
        kwargs['name'] = command['name']
        kwargs['executable'] = command['command']

        if dont_run:
            return ('Monitor', command['arguments'], kwargs)
        else:
            return sys.modules['wine.monitor'].Program(
                command['arguments'],
                **kwargs
            )
    else:
        command_list = [command['command']] + command['arguments']

        if shell_output:
            kwargs['stdout'] = sys.stdout
            kwargs['stderr'] = sys.stderr

        if dont_run:
            return ('Run', command_list, kwargs)
        else:
            return common.Popen(
                command_list,
                **kwargs
            )


def run_old(command, use_url=True, run_in_dir=True, shell_output=None, output_to_files=False, dont_run=False, name=None):
    """
    Intelligently runs a Windows program.
    This function also handles start.exe arguments (like web links), .msi installers and provides
    the alias "cmd" for launching the wineconsole and "winetricks" for running, and optionally downloading
    winetricks, with arguments."""

    if shell_output is True and output_to_files is True:
        raise Exception, "shell_output and output_to_files can't both be True"


    # If command is given as a string
    if type(command) in (str, unicode):
        # Is this is a UNIX path?
        if os.path.lexists(command):
            #print("File exists, using directly")
            _command = [command]
        else:
        # No? Okay, how about a Windows path?
            _command = [util.wintounix(command)]

            if _command[0] == command:
            #    print("File doesn't exist, splitting")
            # Not that either? Okay, it must be a list of commands and/or arguments
            # Convert it to a list of strings
                _command = util.string_split(
                    command,
                    retain_defines=True,
                    remove_escapes=True
                )
            #else:
            #    print("File exists as Windows path, converted")
    else:
        # This is a list, meaning it has to be preparsed, don't change it
        _command = command[:] # make a copy, not a link

    if _command[0] == 'winetricks':
        if dont_run:
            return (
                'winetricks',
                _command[1:], {
                    'shell_output': shell_output,
                    'output_to_files': output_to_files
                }
            )
        else:
            return run_winetricks(
                _command[1:],
                shell_output=shell_output, output_to_files=output_to_files
            )


    # Remove double backslashes
    _command = [ i.replace('\\\\', '\\') for i in _command ]
    # Remove any definition of any WINEPREFIX
    _command = [ i for i in _command if not i.startswith('WINEPREFIX=') ]
    # Remove any 'env' without any definitions after it since that would mess up the command
    if _command[0] == 'env' and '=' not in _command[1]:
        del _command[0]
    if _command[0] == 'wine':
        del _command[0]
    if len(_command) == 1:
        _command[0] = util.string_remove_escapes(_command[0])

    #print("Command split into: {0}".format(_command))
    #return False
    # Get the last executable argument. Used for getting the extension and the path
    last_executable = [ i for i in _command if '.' in i[-5:] ]
    if len(last_executable):
        last_executable = last_executable[-1]
    else:
        last_executable = ''

    # Get any possible path to files or directories
    possible_paths = [ i for i in _command[1:] if not i[0] in ['/','-'] and i[1:].startswith(':\\') ]

    # Get the extension of the last executable argument
    if '.' in last_executable:
        extension = last_executable.split('.')[-1].lower()
    else:
        extension = ''

    if run_in_dir == True:
        #print("Run in dir is True, getting dir of last executable...")
        if last_executable == '':
            #print("Last executable wasn't found, not running in dir.")
            run_in_dir = False
        elif extension != 'url':
            run_in_dir = os.path.dirname(util.wintounix(last_executable))
            #print("Running in path: %s." % run_in_dir)

    # Remove wine command from the string and return the arguments
    #exe_args = util.string_split(exe.replace('wine ',''), remove_escapes=False)

    # Setup the internal variables for how to run the process/program/command
    # Note that _output_to_terminal will be replaced with _output_to_shell later
    _output_to_terminal = None
    _executable = 'wine'
    _arguments = []
    _monitor_process = (shell_output != True)
    _run_as_is = False
    _use_shell = False
    _name = ''

    # Check to see if monitoring was requested off
    for _command_argument in _command:
        if (
            _command_argument.startswith('VINEYARD_NO_MONITORING') and
            _command_argument.split('=')[1].replace('"','').replace("'",'').lower() == 'true'
        ):
            _monitor_process = False

    # Check for command being regedit
    if _command == ['regedit']:
        # Run regedit, without monitoring it
        _output_to_terminal = True
        _arguments = desktop._openargs+['regedit']
        _monitor_process = False
        _name = "Regedit"
    # Check for command being cmd
    elif _command == ['cmd']:
        # Run an interactive command console
        _arguments = desktop._openargs+['wineconsole', '--backend=user', 'cmd']
        _name = "Console"
    elif _command[0] == 'sh':
        _executable = _command[1]
        _arguments = _command[2:]
        _name = "Shell command"
    # Check for control panels
    elif len(_command) == 1 and _command[0].lower().endswith('.cpl'):
        _arguments = desktop._openargs+['control', _command[0]]
        _name = "Control Panel"
    # Check for command being for a URL file
    elif use_url and extension == 'url':
        #print("Looks like a URL file...")
        # This looks like a URL shortcut file, check it to see if it is
        # or if it's just an argument
        #url = re.search(r'(?i).*(\w\:\\.*?.url)["\']?$', exe)
        #if url:
        if last_executable[1:].startswith(':\\'):
            #print("Yup, it's a URL file. Checking if it's a proper URL file...")
            #url_filename = util.wintounix(url.groups()[0])
            url_filename = util.wintounix(last_executable)
            with open(url_filename, 'r') as url_fileobj:
                url = url_fileobj.read()
            url = re.search(r'(?im)\[InternetShortcut\]\s+URL=(.*?)\s+?$', url)
            if url:
                #print("It checks out, run it.")
                # This is a URL, open it in a web browser, of course without monitoring it
                url = url.groups()[0]
                _output_to_terminal = True
                _executable = 'xdg-open'
                _arguments = [url]
                _monitor_process = False
                run_in_path = False
        if not url:
            # This is not a URL, re-run all checks except checking for it being a URL
            return run(command, use_url=False, shell_output=shell_output, dont_run=dont_run)
    # Check for command being a direct URI/URL address
    elif (
        (
            _command[0].lower().endswith('windows\\command\\start.exe') or
            _command[0].lower() == 'start.exe' or
            _command[0].lower() == 'start'
        ) and (
            (
                len(possible_paths) and len(possible_paths[0]) and (
                    os.path.isdir(util.wintounix(possible_paths[0])) or
                    common.any_in_string(['tp://', 'tps://'], possible_paths[0].lower()[:-9])
                )
            ) or (
                '/unix' in [ i.lower() for i in _command ] and
                len(_command) >= 2
            )
        )
    ):
        # This is a command to open a directory or URL, open it using xdg-open,
        # of course without monitoring it
        if '/unix' in [ i.lower() for i in _command ]:
            # This is already a unix path, don't convert
            path = _command[
                [ i.lower() for i in _command ].index('/unix')+1
            ]
            path = util.wintounix(util.unixtowin(util.string_remove_escapes(path)))
        else:
            path = util.wintounix(possible_paths[0])
        _output_to_terminal = True
        _executable = 'xdg-open'
        link = parsers.read_lnk_file(path)
        if link:
            path = link
        _arguments = [path]
        _monitor_process = False
        run_in_path = False
        return run([path], use_url=use_url, shell_output=shell_output, dont_run=dont_run)
    # Check for command being for a program that should be run in a command console
    elif (
        _command[0].lower().endswith('windows\\system32\\cmd.exe') or
        _command[0].lower() == 'cmd.exe' or
        _command[0].lower() == 'cmd'
    ):
        # This is a program that should be run in a command console
        _arguments = desktop._openargs+['wineconsole', '--backend=user', 'cmd']+_command[1:]
        _name = "Console: {0}".format(_command[1])
    # Check for command being an "advanced" call
    # (since we've removed any WINEPREFIX definitions, this has to be for something "fancy")
    elif _command[0] == 'env':
        # This looks like an "advanced" call, run it as it is given
        _run_as_is = True
        _use_shell = True
    # Check for command being for a Windows Installer
    elif extension == 'msi':
        # This is a Windows Installer, run in and monitor it
        _arguments = desktop._openargs+['msiexec', '/i']+_command
        _name = "Installer"
    # Check for command being a .lnk file
    elif extension == 'lnk':
        link = parsers.read_lnk_file(util.wintounix(_command[0]))
        if link:
            # Change the .lnk filename to the one it's pointing to
            # We also convert it to a unix path and back to deal with Wine's
            # inability to handle Windows "wildcards" (the tilde character)
            _command[0] = util.unixtowin(util.wintounix(link))
            # Continue as normal
            _arguments = desktop._openargs+_command
    # Normal Wine command call
    else:
        # This is a general Wine command call, run it as is
        _arguments = desktop._openargs+_command

    if shell_output == True or (_output_to_terminal == True and shell_output != False):
        stdout = sys.stdout
        stderr = sys.stderr
        _output_to_shell = True
    else:
        stdout = subprocess.PIPE
        stderr = subprocess.PIPE
        _output_to_shell = False

    arguments_list = []
    arg_nr = 0
    while arg_nr < len(_arguments):
        # Don't include start.exe command if we're monitoring,
        # it starts a new process and not as a subprocess,
        # so we can't monitor it.
        # The only benefits the start.exe command has are that it
        # can start a program as minimised, maximised or normal
        # - another way to do this would be good so we can translate
        #   the call.
        if (
            _monitor_process and
            (
                _arguments[arg_nr].lower().endswith('windows\\command\\start.exe') or
                _arguments[arg_nr].lower() == 'start.exe' or
                _arguments[arg_nr].lower() == 'start'
            )
        ):
            # Skip arguments to start.exe
            arg_nr += 1
            while _arguments[arg_nr] in ['/M', '/MAX', '/R', '/W', '/Unix', '/L']:
                arg_nr += 1
        else:
            # Convert special Vineyard values in to their real values
            # (CDROM:\\ for example)
            argument = util.enhance_windows_path(
                _arguments[arg_nr]
            )
            arguments_list.append(argument)
            arg_nr += 1

    if _executable != 'wine' and not os.path.lexists(_executable):
        found_executable = sys.modules['wine.command'].which(_executable)
        if found_executable is not None:
            _executable = found_executable

    print(_executable, command)

    try:
        if _monitor_process:
            if name is None and len(_name):
                name = _name

            kwargs = {
                'env': common.ENV,
                'output_to_shell': _output_to_shell,
                'name': name
            }
            if type(run_in_dir) == str:
                kwargs['cwd'] = run_in_dir
            if _run_as_is:
                kwargs['executable'] = None
                if type(command) not in (list, tuple):
                    command = util.string_split(
                        command,
                        retain_defines=True,
                        remove_escapes=True
                    )

                if dont_run:
                    if 'env' in kwargs:
                        kwargs['env'] = {
                            'WINEPREFIX': kwargs['env']['WINEPREFIX'],
                            'HOME': kwargs['env']['HOME']
                        }
                    return ('Monitor', command, kwargs)
                else:
                    return monitor.Program(command, **kwargs)
            else:
                kwargs['executable'] = _executable

                if dont_run:
                    if 'env' in kwargs:
                        kwargs['env'] = {
                            'WINEPREFIX': kwargs['env']['WINEPREFIX'],
                            'HOME': kwargs['env']['HOME']
                        }
                    return ('Monitor', arguments_list, kwargs)
                else:
                    return monitor.Program(arguments_list, **kwargs)
        else:
            if _run_as_is:
                if dont_run:
                    return ('Run', command, {
                        'stdout': stdout,
                        'stderr': stderr,
                        'env': {
                            'WINEPREFIX': common.ENV['WINEPREFIX'],
                            #'HOME': common.ENV['HOME']
                            'XDG_CONFIG_HOME': common.ENV['XDG_CONFIG_HOME'],
                            'XDG_DATA_HOME': common.ENV['XDG_DATA_HOME']
                        }
                    })
                else:
                    return common.Popen(command,
                        stdout=stdout, stderr=stderr,
                        env=common.ENV, shell=_use_shell
                    )
            else:
                if dont_run:
                    return ('Run', [_executable]+arguments_list, {
                        'stdout': stdout,
                        'stderr': stderr,
                        'env': {
                            'WINEPREFIX': common.ENV['WINEPREFIX'],
                            #'HOME': common.ENV['HOME']
                            'XDG_CONFIG_HOME': common.ENV['XDG_CONFIG_HOME'],
                            'XDG_DATA_HOME': common.ENV['XDG_DATA_HOME']
                        }
                    })
                else:
                    return common.Popen([_executable]+arguments_list,
                        stdout=stdout, stderr=stderr,
                        env=common.ENV, shell=_use_shell
                    )

    except OSError:
        if len(_executable):
            print("Couldn't run command: %s" % str([_executable]+arguments_list))
        else:
            print("Couldn't run command: %s" % arguments_list)
        return False

